@@ -1,19 +1,19 @@
 import datetime
 from StringIO import StringIO
+import django
 from django.conf import settings
 from django.core.exceptions import ImproperlyConfigured
 from django.core.serializers import json
-from django.template import loader, Context
 from django.utils import simplejson
 from django.utils.encoding import force_unicode
 from tastypie.bundle import Bundle
-from tastypie.exceptions import UnsupportedFormat
-from tastypie.utils import format_datetime, format_date, format_time
-from tastypie.fields import ApiField, ToOneField, ToManyField
+from tastypie.exceptions import BadRequest, UnsupportedFormat
+from tastypie.utils import format_datetime, format_date, format_time, make_naive
 try:
-    import lxml
-    from lxml.etree import parse as parse_xml
-    from lxml.etree import Element, tostring
+    import defusedxml.lxml as lxml
+    from defusedxml.common import DefusedXmlException
+    from defusedxml.lxml import parse as parse_xml
+    from lxml.etree import Element, tostring, LxmlError
 except ImportError:
     lxml = None
 try:
@@ -21,6 +21,38 @@ try:
     from django.core.serializers import pyyaml
 except ImportError:
     yaml = None
+try:
+    import biplist
+except ImportError:
+    biplist = None
+
+
+# Ugh & blah.
+# So doing a regular dump is generally fine, since Tastypie doesn't usually
+# serialize advanced types. *HOWEVER*, it will dump out Python Unicode strings
+# as a custom YAML tag, which of course ``yaml.safe_load`` can't handle.
+if yaml is not None:
+    from yaml.constructor import SafeConstructor
+    from yaml.loader import Reader, Scanner, Parser, Composer, Resolver
+
+    class TastypieConstructor(SafeConstructor):
+        def construct_yaml_unicode_dammit(self, node):
+            value = self.construct_scalar(node)
+            try:
+                return value.encode('ascii')
+            except UnicodeEncodeError:
+                return value
+
+    TastypieConstructor.add_constructor(u'tag:yaml.org,2002:python/unicode', TastypieConstructor.construct_yaml_unicode_dammit)
+
+    class TastypieLoader(Reader, Scanner, Parser, Composer, TastypieConstructor, Resolver):
+        def __init__(self, stream):
+            Reader.__init__(self, stream)
+            Scanner.__init__(self)
+            Parser.__init__(self)
+            Composer.__init__(self)
+            TastypieConstructor.__init__(self)
+            Resolver.__init__(self)
 
 
 class Serializer(object):
@@ -30,36 +62,47 @@ class Serializer(object):
     This handles most types of data as well as the following output formats::
 
         * json
-        * jsonp
+        * jsonp (Disabled by default)
         * xml
         * yaml
         * html
+        * plist (see http://explorapp.com/biplist/)
 
     It was designed to make changing behavior easy, either by overridding the
     various format methods (i.e. ``to_json``), by changing the
     ``formats/content_types`` options or by altering the other hook methods.
     """
-    formats = ['json', 'jsonp', 'xml', 'yaml', 'html']
-    content_types = {
-        'json': 'application/json',
-        'jsonp': 'text/javascript',
-        'xml': 'application/xml',
-        'yaml': 'text/yaml',
-        'html': 'text/html',
-    }
+
+    formats = ['json', 'xml', 'yaml', 'html', 'plist']
+
+    content_types = {'json': 'application/json',
+                     'jsonp': 'text/javascript',
+                     'xml': 'application/xml',
+                     'yaml': 'text/yaml',
+                     'html': 'text/html',
+                     'plist': 'application/x-plist'}
 
     def __init__(self, formats=None, content_types=None, datetime_formatting=None):
-        self.supported_formats = []
-        self.datetime_formatting = getattr(settings, 'TASTYPIE_DATETIME_FORMATTING', 'iso-8601')
+        if datetime_formatting is not None:
+            self.datetime_formatting = datetime_formatting
+        else:
+            self.datetime_formatting = getattr(settings, 'TASTYPIE_DATETIME_FORMATTING', 'iso-8601')
 
-        if formats is not None:
-            self.formats = formats
+        self.supported_formats = []
 
         if content_types is not None:
             self.content_types = content_types
 
-        if datetime_formatting is not None:
-            self.datetime_formatting = datetime_formatting
+        if formats is not None:
+            self.formats = formats
+
+        if self.formats is Serializer.formats and hasattr(settings, 'TASTYPIE_DEFAULT_FORMATS'):
+            # We want TASTYPIE_DEFAULT_FORMATS to override unmodified defaults but not intentational changes
+            # on Serializer subclasses:
+            self.formats = settings.TASTYPIE_DEFAULT_FORMATS
+
+        if not isinstance(self.formats, (list, tuple)):
+            raise ImproperlyConfigured('Formats should be a list or tuple, not %r' % self.formats)
 
         for format in self.formats:
             try:
@@ -88,6 +131,7 @@ class Serializer(object):
 
         Default is ``iso-8601``, which looks like "2010-12-16T03:02:14".
         """
+        data = make_naive(data)
         if self.datetime_formatting == 'rfc-2822':
             return format_datetime(data)
 
@@ -175,13 +219,13 @@ class Serializer(object):
             return dict((key, self.to_simple(val, options)) for (key, val) in data.iteritems())
         elif isinstance(data, Bundle):
             return dict((key, self.to_simple(val, options)) for (key, val) in data.data.iteritems())
-        elif isinstance(data, ApiField):
-            if isinstance(data, ToOneField):
+        elif hasattr(data, 'dehydrated_type'):
+            if getattr(data, 'dehydrated_type', None) == 'related' and data.is_m2m == False:
                 if data.full:
                     return self.to_simple(data.fk_resource, options)
                 else:
                     return self.to_simple(data.value, options)
-            elif isinstance(data, ToManyField):
+            elif getattr(data, 'dehydrated_type', None) == 'related' and data.is_m2m == True:
                 if data.full:
                     return [self.to_simple(bundle, options) for bundle in data.m2m_bundles]
                 else:
@@ -229,13 +273,13 @@ class Serializer(object):
             element = Element(name or 'object')
             for field_name, field_object in data.data.items():
                 element.append(self.to_etree(field_object, options, name=field_name, depth=depth+1))
-        elif isinstance(data, ApiField):
-            if isinstance(data, ToOneField):
+        elif hasattr(data, 'dehydrated_type'):
+            if getattr(data, 'dehydrated_type', None) == 'related' and data.is_m2m == False:
                 if data.full:
                     return self.to_etree(data.fk_resource, options, name, depth+1)
                 else:
                     return self.to_etree(data.value, options, name, depth+1)
-            elif isinstance(data, ToManyField):
+            elif getattr(data, 'dehydrated_type', None) == 'related' and data.is_m2m == True:
                 if data.full:
                     element = Element(name or 'objects')
                     for bundle in data.m2m_bundles:
@@ -250,10 +294,16 @@ class Serializer(object):
             element = Element(name or 'value')
             simple_data = self.to_simple(data, options)
             data_type = get_type_string(simple_data)
+
             if data_type != 'string':
                 element.set('type', get_type_string(simple_data))
+
             if data_type != 'null':
-                element.text = force_unicode(simple_data)
+                if isinstance(simple_data, unicode):
+                    element.text = simple_data
+                else:
+                    element.text = force_unicode(simple_data)
+
         return element
 
     def from_etree(self, data):
@@ -296,7 +346,11 @@ class Serializer(object):
         """
         options = options or {}
         data = self.to_simple(data, options)
-        return simplejson.dumps(data, cls=json.DjangoJSONEncoder, sort_keys=True)
+
+        if django.get_version() >= '1.5':
+            return json.json.dumps(data, cls=json.DjangoJSONEncoder, sort_keys=True, ensure_ascii=False)
+        else:
+            return simplejson.dumps(data, cls=json.DjangoJSONEncoder, sort_keys=True, ensure_ascii=False)
 
     def from_json(self, content):
         """
@@ -308,9 +362,16 @@ class Serializer(object):
         """
         Given some Python data, produces JSON output wrapped in the provided
         callback.
+
+        Due to a difference between JSON and Javascript, two
+        newline characters, \u2028 and \u2029, need to be escaped.
+        See http://timelessrepo.com/json-isnt-a-javascript-subset for
+        details.
         """
         options = options or {}
-        return '%s(%s)' % (options['callback'], self.to_json(data, options))
+        json = self.to_json(data, options)
+        json = json.replace(u'\u2028', u'\\u2028').replace(u'\u2029', u'\\u2029')
+        return u'%s(%s)' % (options['callback'], json)
 
     def to_xml(self, data, options=None):
         """
@@ -323,14 +384,24 @@ class Serializer(object):
 
         return tostring(self.to_etree(data, options), xml_declaration=True, encoding='utf-8')
 
-    def from_xml(self, content):
+    def from_xml(self, content, forbid_dtd=True, forbid_entities=True):
         """
         Given some XML data, returns a Python dictionary of the decoded data.
+
+        By default XML entity declarations and DTDs will raise a BadRequest
+        exception content but subclasses may choose to override this if
+        necessary.
         """
         if lxml is None:
             raise ImproperlyConfigured("Usage of the XML aspects requires lxml.")
 
-        return self.from_etree(parse_xml(StringIO(content)).getroot())
+        try:
+            parsed = parse_xml(StringIO(content), forbid_dtd=forbid_dtd,
+                               forbid_entities=forbid_entities)
+        except (LxmlError, DefusedXmlException):
+            raise BadRequest
+
+        return self.from_etree(parsed.getroot())
 
     def to_yaml(self, data, options=None):
         """
@@ -350,7 +421,27 @@ class Serializer(object):
         if yaml is None:
             raise ImproperlyConfigured("Usage of the YAML aspects requires yaml.")
 
-        return yaml.safe_load(content)
+        return yaml.load(content, Loader=TastypieLoader)
+
+    def to_plist(self, data, options=None):
+        """
+        Given some Python data, produces binary plist output.
+        """
+        options = options or {}
+
+        if biplist is None:
+            raise ImproperlyConfigured("Usage of the plist aspects requires biplist.")
+
+        return biplist.writePlistToString(self.to_simple(data, options))
+
+    def from_plist(self, content):
+        """
+        Given some binary plist data, returns a Python dictionary of the decoded data.
+        """
+        if biplist is None:
+            raise ImproperlyConfigured("Usage of the plist aspects requires biplist.")
+
+        return biplist.readPlistFromString(content)
 
     def to_html(self, data, options=None):
         """
